#!/usr/bin/perl
$prog=$0;
$prog =~ s/.*\///;

# block by block compare, sync and count
$bsz=65536;
$pflg=0;
$vflg=0;
$mode="<";	# default mode is just read
$usage="$prog [-b <block size>[K|M]] [-{r|w}] [-p] [-v] <source-file> <destination-file>
	-b	size of blocks read/written (default $bsz)
	-r	read only mode (default)
	-w	write differences (actually sync the two files)
	-p	update permissions (mode, owner, group, times)
	-v	verbose
Note: destination will be created if necessary
Source and destination must be files (no directories, no multiple files)";

# args
while ($#ARGV >= 0) {
	if ($ARGV[0] eq "-b") {
		shift @ARGV;
		$bsz=$ARGV[0];
		$bsz =~ y/A-Z/a-z/;
		$e=substr($bsz, length($bsz)-1);
		if ($e eq "k") {
			$bsz=substr($bsz, 0, length($bsz)-1);
			$bsz *= 1024;
		}
		if ($e eq "m") {
			$bsz=substr($bsz, 0, length($bsz)-1);
			$bsz *= 1024*1024;
		}
		if ($bsz < 0) {
			printf STDERR "%s: Usage %s\n", $prog, $usage;
			exit 1;
		}
		shift @ARGV;
		next;
	}
	if ($ARGV[0] eq "-r") {
		shift @ARGV;
		$mode="<";
		next;
	}
	if ($ARGV[0] eq "-w") {
		shift @ARGV;
		$mode="+<";
		next;
	}
	if ($ARGV[0] eq "-p") {
		shift @ARGV;
		$pflg=1;
		next;
	}
	if ($ARGV[0] eq "-v") {
		shift @ARGV;
		$vflg=1;
		$est=0;
		$| = 1;
		next;
	}
	last;
}

if ($#ARGV != 1) {
	printf STDERR "%s: Usage %s\n", $prog, $usage;
	exit 1;
}
$sfile=$ARGV[0];
unless (open(S, $sfile)) {
	printf STDERR "%s: open %s: %s", $prog, $sfile, $! . "\n";
	exit 1;
}
binmode S;
if ($vflg == 1) {
	@stinfo=stat($sfile);
	$est=int(($stinfo[7]-1)/$bsz)+1;
}
$dfile=$ARGV[1];
if ($mode eq "+<" && ! -e $dfile) {
	$mode="+>";
}
unless (open(D, $mode . $dfile)) {
	printf STDERR "%s: open %s: %s", $prog, $dfile, $! . "\n";
	exit 1;
}
binmode D;

$dlen=$bsz;
$cnt=0;
$grow=0;
$tot=0;
while (1) {
	$slen=sysread(S, $sbuf, $bsz);
	if ($slen < 0) {
		printf STDERR "%s: read %s: %s", $prog, $sfile, $! . "\n";
		exit 1;
	}
	if ($slen == 0) {
		last;
	}
	$tot++;
	if ($dlen) {
		$dlen=sysread(D, $dbuf, $slen);
		if ($dlen < 0) {
			printf STDERR "%s: read %s: %s", $prog, $dfile, $! . "\n";
			exit 1;
		}
	}
	if (!$dlen || $sbuf ne $dbuf) {
		if ($mode ne "<") {
			if ($dlen) {
				seek(D, -$dlen, 1);
			}
			$wlen=syswrite(D, $sbuf, $slen);
			if ($wlen != $slen) {
				printf STDERR "%s: write %s: %s", $prog, $dfile, $! . "\n";
				exit 1;
			}
		}
		if ($dlen) {
			$cnt++;
		} else {
			$grow++;
		}
	}
	if ($vflg) {
		printf "%d+%d/%d/%d\r", $cnt, $grow, $tot, $est;
	}
}
close S;
close D;
printf "Difference %d/%d blocks", $cnt, $tot;
if ($grow) {
	printf " + grow by %d blocks", $grow;
}
@st=stat($sfile);
@dt=stat($dfile);
if ($st[7] < $dt[7]) {
	if ($mode ne "<") {
		unless (truncate($dfile, $st[7])) {
			$err=$! . "\n";
			printf "\n";
			printf STDERR "%s: truncate %s: %s", $prog, $dfile, $err;
			exit 1;
		}
	}
	printf " + truncate\n";
} else {
	printf "\n";
}
if ($pflg && $mode ne "<") {
	if ($st[2] != $dt[2]) {
		unless (chmod($st[2], $dfile)) {
			$err=$! . "\n";
			printf STDERR "%s: chmod %s: %s", $prog, $dfile, $err;
		}
	}
	if ($st[9] != $dt[9]) {
		unless (utime($st[8], $st[9], $dfile)) {
			$err=$! . "\n";
			printf STDERR "%s: utime %s: %s", $prog, $dfile, $err;
		}
	}
	if ($st[4] != $dt[4] || $st[5] != $dt[5]) {
		unless (chown($st[4], $st[5], $dfile)) {
			$err=$! . "\n";
			printf STDERR "%s: chown %s: %s", $prog, $dfile, $err;
		}
	}
}
exit 0;
